#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <demo-mod.h>
#include <stdlib.h>
#include <string.h>

typedef struct dynmod {
  const char *name;
  int do_msg;
  lt_dlhandle handle;
  get_proc *get;
  set_proc *set;
  str_proc *str;
} dynmod;

static dynmod dynmod1 = { "mod1", 1, NULL, NULL, NULL, NULL };
static dynmod dynmod2 = { "mod2", 1, NULL, NULL, NULL, NULL };

static void dynclose(dynmod *d) {
  if (d->handle == NULL) {
    printf("dynamic module %s not open\n", d->name);
    return;
  }

  lt_dlclose(d->handle);
  d->handle = NULL;
  d->get = NULL;
  d->set = NULL;
  d->str = NULL;
  lt_dlerror();	/* Clear any existing error */
}

/* Try to get symbols */
static int dynfuncs(dynmod *d, const char *name) {
  size_t l;
  char *s;
  int res;
  init_proc *init;

  if (d->do_msg)
    printf("%s to open %s\n", d->handle == NULL ? "Failed" : "Succeeded", name);
  if (d->handle == NULL)
    return 0;

  res = 1;
  l = strlen(d->name)+5;
  s = (char *) malloc(l+5);
  strcpy(s, d->name);
  strcat(s, "_LTX_");

#define GET_PTR(str, ptr) \
  strcpy(s+l, str); \
  *(void **)(&ptr) = lt_dlsym(d->handle, s); \
  if (ptr == NULL) \
    res = 0
#define GET_SYM(name) \
  GET_PTR(#name, d->name)
#define DO_INIT \
  GET_PTR("init", init); \
  else \
    init()

  GET_SYM(get);
  GET_SYM(set);
  GET_SYM(str);
  DO_INIT;

#undef DO_INIT
#undef GET_SYM
#undef DO_INIT

  if (res == 0 && d->do_msg)
    printf("Failed to resolve symbols from %s\n", name);
  else
    d->do_msg = 0;
  free(s);
  return res;
}

static void dynopen(dynmod *d) {
  char *s;
  int l;

  if (d->handle != NULL) {
    printf("dynamic module %s already open\n", d->name);
    return;
  }

#if 0

  /* First try to append ".a" to the module name,
   * in case the modules are preloaded (via libtool)
   */
  
  l = strlen(d->name);
  s = (char *)malloc(l+3);
  strcpy(s, d->name);
  strcpy(s+l, ".a");
  d->handle = lt_dlopen(s);
  {
    int res;

    res = dynfuncs(d, s);
    free(s);
    if (res)
      return;
  }

#endif

  /* Then try bare module name to be located via MODULE_PATH and/or LD_LIBRARY_PATH */
  d->handle = lt_dlopenext(d->name);
  if (dynfuncs(d, d->name))
    return;

  printf("Unable to load module %s\n", d->name);
  exit(1);
}

static int dynrun(dynmod *d) {
  int i, j;

  i = d->get();
  d->set(i+5);
  j = d->get();
  printf("Module %s (%s) %d+5=%d\n", d->name, d->str(), i, j);
  return j;
}

int main(int argc, char **argv) {
  int res;
  const char *mod_path;

  printf("Entering app1\n");

  LTDL_SET_PRELOADED_SYMBOLS();

  lt_dlinit();

  mod_path = getenv("MODULE_PATH");
  if (mod_path == NULL)
    mod_path = MODDIR;
  lt_dlsetsearchpath(mod_path);

  dynopen(&dynmod1);		/* val1 = 17 */
  dynrun(&dynmod1);		/* val1 = 22 */
  dynopen(&dynmod2);		/* val2 = 5 */
  dynrun(&dynmod2);		/* val2 = 10 */
  dynrun(&dynmod1);		/* val1 = 27 */
  dynrun(&dynmod2);		/* val2 = 15 */
  dynclose(&dynmod1);
  dynopen(&dynmod1);		/* val1 = 17 */
  dynrun(&dynmod1);		/* val1 = 22 */
  dynrun(&dynmod2);		/* val2 = 27 */
  dynclose(&dynmod2);
  dynopen(&dynmod2);		/* val2 = 5 */
  res = dynrun(&dynmod1);	/* val1 = 27, res = 27 */
  res += dynrun(&dynmod2);	/* val2 = 10, res = 37 */
  dynclose(&dynmod1);
  dynclose(&dynmod2);
  
  lt_dlexit();

  printf("Leaving app1: res=%d\n", res);
  return !(res == 37);
}
